import Prelude hiding (catch)
import Control.Exception hiding(bracket)
import Char
import Text.ParserCombinators.HuttonMeijer hiding( ident, skip)
import Control.Monad
import Control.Monad.State
import Data.Maybe
import Debug.Trace
import Data.List hiding( partition )
import Debug.Trace

at x y = x !! y
apply f a = f a
extend v a r = \x-> if x==v then a else r x

infinite x = x : infinite x

type X = String
type M = String
type C = String
type Loc = Int
type Bindings = [(X, EV)]
type FieldDecls = [X]
type Formals = [X]
type DynamicFields = Formals

type Methods = [Method]
type Classes = [Class]
type Args = [E]
type NameMap = [(NameKey,(Int, [C]))] -- the generated name for (C,Bindings) is C++Int
data NameKey = ClassKey C [EV] | MethodKey M Int [EV] -- Int is the number of args

data Primitive = Null | StringVal String | IntVal Int | BoolVal Bool | ListVal [V]
data V = Prim Primitive | Obj C Env | Void

type Prog = Classes

data Class = ClassDef C Formals FieldDecls E Methods

data Modifier = ClassMod | InstanceMod
data Method = MethodDef Modifier M Formals E
    deriving Eq

data Op = Add | Sub | Mul | Div | Eq | NEq | Lt | Gt | Mod
    deriving Eq

data E = Value V       {- \quad\quad\quad{\it constant value} -}
    | Var X            {- \quad\quad\quad{\it variable} -}
    | ClassVal C       {- \quad\quad\quad{\it class name} -}
    | Let X E E        {- \quad\quad\quad{\it variable declaration} -}
    | Assign X E       {- \quad\quad\quad{\it assignment} -}
    | Seq E E          {- \quad\quad\quad{\it sequence} -}
    | Bop Op E E       {- \quad\quad\quad{\it binary operator} -}
    | If E E E         {- \quad\quad\quad{\it conditional} -}
    | While E E        {- \quad\quad\quad{\it iteration} -}
    | Call E M Args    {- \quad\quad\quad{\it method invocation} -}
    | Invoke E E Args  {- \quad\quad\quad{\it reflective method invocation} -}
    | New C Args       {- \quad\quad\quad{\it constructor call} -}
    | CT E E           {- \quad\quad\quad{\it execute at compile-time } -}
    | RT E             {- \quad\quad\quad{\it execute at runtime } -}
    | IsCT E           {- \quad\quad\quad{\it tests for a compile-time value } -}
    deriving Eq

type Env = [(X, Loc)]
type Store = [(Loc, EV)]
feval :: Env -> E -> State (Prog, Store, NameMap) V
feval_fun (Value v) env = return v
env_fun x = 0 -- dummy
store_fun l = (Con (Prim Null)) -- dummy

data EV = Abs V | Con V | Top | Bottom
env_fun :: X -> Loc
store_fun :: Loc -> EV
feval_fun :: E -> Env -> State (Prog, Store, NameMap) V

feval env (Value v) = return v

feval env (Bop op e1 e2) = do
  Prim v1 <- feval env e1
  Prim v2 <- feval env e2
  return (Prim (compute op v1 v2))
feval env (Var x) = do
  (_,st,_) <- get
  return (getValue (rlookup (rlookup x env) st))
feval env (Let x e1 e2) = do
  v <- feval env e1
  varx <- allocate (bind x (Con v))
  feval (varx++env) e2
feval env (Assign x e) = do
  v <- feval env e
  update (rlookup x env) (Con v)
  return v
feval env (Seq e1 e2) = do
  feval env e1
  feval env e2
feval env (If e1 e2 e3) = do
  b <- feval env e1
  case b of
    Prim (BoolVal True) -> feval env e2
    Prim (BoolVal False) -> feval env e3
feval env (While e1 e2) = do
  feval env (If e1 (Seq e2 (While e1 e2)) (Value (Prim Null)))
feval env (Call e m args) = do
  Obj c env' <- feval env e
  vals <- mapM (feval env) args
  MethodDef _ _ xs body <- findMethod c m (length args)
  varsx <- allocate (zip xs (toCon vals))
  varthis <- allocate (bind "this" (Con (Obj c env')))
  feval (varthis++varsx++env') body
feval env (Invoke e em args) = do
  Prim (StringVal m) <- feval env em
  feval env (Call e m args)
feval env (New c args) = do
  ClassDef dummyC xs fs ctor _ <- findClass c
  vals <- mapM (feval env) args
  varsx <- allocate (zip xs (toCon vals))
  env' <- allocate (zip fs (infinite Bottom))
  feval (varsx++env') ctor
  return (Obj c env')

feval env (CT e e') = do
  feval env e
feval env (IsCT e) = do
  return (Prim (BoolVal True))
feval env (RT e) = do
  feval env e


data PEResult = R E EV

peval_fun :: E -> Env -> State (Prog, Store, NameMap) PEResult
instance Show PEResult where
  show (R e ev) = "("++show e++", "++show ev++")"

peval :: Env -> E -> State (Prog, Store, NameMap) PEResult
peval_fun (Value v) env = return (mCon v)

peval env exp = do
    (program,st,_) <- get
    let label = show exp
    r <- peval1 env (trace ("peval " ++ show (envStore env st) ++ " " ++ label) exp)
    return $ trace ("***[" ++ trim label ++ "]= " ++ show r) r


peval1 env (Value v) = return (mAbs (Value v) (Abs v))


peval1 env (Var x) = do
  (_,st, _) <- get
  case rlookup (rlookup x env) st of
    Con v -> return (mCon v)
    p -> return (mAbs (Var x) p)

peval1 env (Let x e1 e2) = do
      R e1' p1 <- peval env e1
      varx <- allocate (bind x p1)
      R e2' p2 <- peval (varx++env) e2
      case p1 of
        Con v -> return (R e2' p2)
        otherwise -> 
         return (mAbs (Let x e1' e2') p2)

peval1 env (CT e e') = do
  v' <- feval env e'
  if v' == Prim (BoolVal True) then do
    v <- feval env e
    return (mCon v)
  else peval env e

peval1 env (IsCT e) = do
  R e' p <- peval env e
  case p of
    Con v -> return (mCon (Prim (BoolVal True)))
    otherwise -> return (mCon (Prim (BoolVal False)))

peval1 env (RT e) = do
  return (mAbs e Top)

peval1 env (Bop op e1 e2) = do
  R e1' p1 <- peval env e1
  R e2' p2 <- peval env e2
  case (p1, p2) of
    (Con (Prim v1), Con (Prim v2)) -> 
      let v = compute op v1 v2  in return (mCon (Prim v))
    (Con (Prim v1), Abs (Prim v2)) -> 
      let v = compute op v1 v2  in return (mCon (Prim v))
    (Abs (Prim v1), Con (Prim v2)) -> 
      let v = compute op v1 v2  in return (mCon (Prim v))
    (Abs (Prim v1), Abs (Prim v2)) -> 
      let v = compute op v1 v2  in return (mAbs (Value (Prim v)) (Abs (Prim v)))
    otherwise -> return (mAbs (Bop op e1' e2') Top)

peval1 env (Assign x e) = do
  (_,st,_) <- get
  case fromJust(lookup (rlookup x env) st) of
    Con v -> do
      v' <- feval env e
      update (rlookup x env) (Con v')
      return (mCon v')
    Abs v -> do
      R e' p <- peval env e
      update (rlookup x env) (toAbs p)
      return (mAbs (Assign x e') p)
    Top -> do
      R e' p <- peval env e
      return (mAbs (Assign x e') Top)      
    Bottom -> do
      R e' p <- peval env e
      update (rlookup x env) p
      case p of
        Con v' -> do
          return (mCon v')
        _ -> do
          return (mAbs (Assign x e') p)

peval1 env (Seq e1 e2) = do
  R e1' p1 <- peval env e1
  R e2' p2 <- peval env e2
  return (mAbs (makeSeq e1' e2') p2)

peval1 env (If e1 e2 e3) = do
  R e1' p1 <- peval env e1
  case p1 of
    Con (Prim (BoolVal True)) -> peval env e2
    Con (Prim (BoolVal False)) -> peval env e3
    otherwise -> checkStore env (If e1') e2 e3

peval1 env (While e1 e2) = do
  R e1' p1 <- peval env e1
  case p1 of
    Con (Prim (BoolVal True)) -> peval env (Seq e2 (While e1 e2))
    Con (Prim (BoolVal False)) -> peval env (Value (Prim Null))
    otherwise -> checkStore env (While) e1 e2

peval1 env (New c args) = do
  argResults <- mapM (peval env) args
  if any isCompileTime (projectValue argResults) then do
    (memoized, c', ClassDef _ xs fs ctor methods) <- findMemoClass c (projectValue argResults)
    varsx <- allocate (zip xs (projectValue argResults))
    let varsxd = getRuntimeNames xs argResults
    let argsd = getRuntimeExprs argResults
    env' <- allocate (zip fs (infinite Bottom))
    varthis <- allocate (bind "this" (Abs (Obj c' env')))
    R ctor' _ <- peval (varthis++varsx++env') ctor
    when(not memoized) (do
      methods' <- mapM (pevalm (varthis++env')) methods
      fs' <- getRuntimeFields fs env'
      addClass (ClassDef c' varsxd fs' ctor' methods'))
    return (mAbs (New c' argsd) (Abs (Obj c' env')))
  else do
    ClassDef _ _ fs _ _ <- findClass c
    env' <- allocate (zip fs (infinite Top))
    return (mAbs (New c (getExprs argResults)) (Abs (Obj c env')))

peval1 env (Call e m args) = do
  R e' p <- peval env e
  argResults <- mapM (peval env) args
  case p of
    Con (Obj c env') -> do
      if all isCompileTime (projectValue argResults) then do
        MethodDef _ _ xs body <- findMethod c m (length args)
        varsx <- allocate (zip xs (projectValue argResults))
        varthis <- allocate (bind "this" (Con (Obj c env')))
        v <- feval (varthis++varsx++env') body
        return (mCon v)
      else 
        doCall ClassMod env' c (ClassVal c) m argResults
    Abs (Obj c env') -> -- target is an approximate object
      if any isCompileTime (projectValue argResults) then do
        varthis <- allocate (bind "this" (Abs (Obj c env')))
        doCall InstanceMod (varthis++env') c e' m argResults
      else 
        return (mAbs (Call e' m (getExprs argResults)) Top)
    otherwise -> --target is unknown
      if any isCompileTime (projectValue argResults) then do
        m' <- specializeAll env m argResults
        let argsd = getRuntimeExprs argResults
        return (mAbs (Call e' m' argsd) Top)
      else 
        return (mAbs (Call e' m (getExprs argResults)) Top)

peval1 env (Invoke e em args) = do
  R em' p <- peval env em
  case em' of
    Value (Prim (StringVal m)) -> peval env (Call e m args)
    otherwise -> do
      R e' p' <- peval1 env e
      argResults <- mapM (peval env) args
      return (mAbs (Invoke e' em' (getExprs argResults)) Top)

--pevalm env meth | trace ("pevalM " ++ show meth) False = undefined
traceMeth m = trace ("NEW METHOD " ++ show m) m

pevalm env (MethodDef modifier m xs e) = do
  env' <- allocate [(x, Top) | x <- xs]
  R e' _ <- peval (env++env') e
  return (traceMeth (MethodDef modifier m xs e'))


doCall modifier env c e m argResults = do
  (memoized, m', MethodDef _ _ xs body) <- findMemoCall c m (projectValue argResults)
  let argsd = getRuntimeExprs argResults
  when(not memoized) ( do
      varsx <- allocate (zip xs (projectValue argResults))
      R body' p <- peval (varsx++env) body
      let varsxd = getRuntimeNames xs argResults
      addMethod c (MethodDef modifier m' varsxd body'))
  return (mAbs (Call e m' argsd) Top)


instance Show Modifier where
  show(ClassMod) = "static"
  show(InstanceMod) = "method"

{-
Null | StringVal String | IntVal Int | BoolVal Bool
        | ListVal [V] | Obj C Loc Env DynamicFields | Void
-}

instance Show Primitive where
  show Null = "null"
  show (StringVal s) = "\""++s++"\""
  show (IntVal i) = show i
  show (BoolVal b) = show b
  show (ListVal l) = show l

instance Show V where
  show (Prim p) = show p
  show (Obj c env) =  c++show env
  show Void = "()"

instance Show EV where
  show (Abs v) = "~"++show v
  show (Con v) = show v
  show Bottom = "?"
  show Top = "!"

instance Show Class where
  show (ClassDef c xs bindings ctor methods) =
      "class "++c++ showFormals xs++" {\n"++
        showVars bindings ++
        (if ctor==(Value (Prim Null)) then "" else "init {\n" ++ show ctor++";\n"++"}\n") ++
        showseq "\n" methods ++
        "}"

showFormals f = "("++ intercalate ", " f ++ ")"
showVars [] = ""
showVars (v:vs) = "var "++v++";\n"++(showVars vs)


instance Show Method where
  show (MethodDef s x xs e) = show s++" "++x++ showFormals xs++" {\n"++show e++";\n}"

instance Show Op where
  show(Add) = "+"
  show(Mul) = "*"
  show(Sub) = "-"
  show(Div) = "/"
  show(Eq) = "="
  show(NEq) = "!="
  show(Lt) = "<"
  show(Gt) = ">"
  show(Mod) = "%"

instance Show E where
  show (Bop o e1 e2) = show e1++show o++show e2
  show (Var a) = a
  show (Value v) = show v
  show (Assign x a) = x ++ " := " ++ show a
  show (Let v e b) = "var " ++v++" = "++show e++ "; \n"++ show b
  show (If a b c) = "if ("++show a++ ") {\n"++ show b++ ";\n}\nelse {\n"++ show c++";\n}"
  show (Seq (a) (b)) = show a ++";\n"++ show b
  show (While a b) = "while ("++ show a++ ") {\n"++ show b++"\n}"
  show (ClassVal c) = c
  show (Call (e) m a) = show e++ "."++ m++ "(" ++ showlist ", " a++ ")"
  show (New c a) = "new "++c++ "("++ showlist ", " a++ ")"
  show (Invoke a m p) = show a++ "@"++ show m++ "("++ showlist ", " p ++")"
  show (CT e c) = "CT("++show e++", "++show c++")"
  show (RT e) = "RT("++show e++")"
  show (IsCT e) = "IsCT("++show e++")"


showlist sep [] = ""
showlist sep [x] = show x
showlist sep (x:xs) = show x++sep++showlist sep xs
showseq sep [] = ""
showseq sep (x:xs) = show x++sep++showseq sep xs

update l v = do
    (p,m,n) <- get
    case find (\(l1, _) -> l1 == l) m of
      Just (a, b) -> do
        put (p, (l,v):(delete (a, b) m), n)
      otherwise ->
        put (p, (l,v):m, n)

allocate1 (x,v) = do
    (p,m, n) <- get
    let l = length m
    put (p, (l,v):m, n)
    return (x,l)

allocate :: [(X, EV)] -> State (Prog, Store, NameMap) Env
allocate statEnv = mapM allocate1 statEnv

findMemoCall c m vargs = do
  mdef <- findMethod c m (length vargs)
  (p,st,n) <- get
  case lookup (MethodKey m (length vargs) vargs) n of
    Just (l, cs) -> do
      if (elem c cs) then 
        return (True, m++"$"++show l, mdef)
       else do
        put (p,st,(((MethodKey m (length vargs) vargs),(l, c:cs)):n))
        return (False, m++"$"++show l, mdef)
    Nothing -> do
      let l = (length n) + 1
      put (p,st,(((MethodKey m (length vargs) vargs),(l, [c])):n))
      return (False, m++"$"++show l, mdef)

findMemoClass c statEnv = do
  (p,st,n) <- get
  case lookup (ClassKey c statEnv) n of
    Just (l, [c]) -> do
      cdef <- findClass (c++"$"++show l)
      return (True, c++"$"++show l, cdef)
    Nothing -> do
      let l = (length n) + 1
      put (p,st,(((ClassKey c statEnv),(l, [c])):n))
      cdef <- findClass c
      return (False, c++"$"++show l, cdef)


deepEqV :: V -> V -> Store -> Store ->  State (Prog, Store, NameMap) Bool
deepEqV v1 v2 s1 s2 =
  case (v1, v2) of
    (Prim p1, Prim p2) -> deepEqPrim p1 p2 s1 s2
    (Obj c1 e1, Obj c2 e2) -> do
      b <- (deepEqEnv e2 s1 s2)
      return ((c1==c2) && (e1==e2) && b)
    (Void, Void) -> return True
    (_, _) -> return False

deepEqPrim p1 p2 s1 s2 =
  case (p1, p2) of
    (Null, Null) -> return True
    (StringVal s1, StringVal s2) -> return (s1 == s2)
    (IntVal i1, IntVal i2) -> return (i1 == i2)
    (BoolVal b1, BoolVal b2) -> return (b1 == b2)
    (ListVal l1, ListVal l2) -> return (l1 == l2)
    (_, _) -> return False

eqLoc :: Loc -> Store -> Store -> State (Prog, Store, NameMap) Bool
eqLoc loc s1 s2 = do
  let ev1 = rlookup loc s1
  let ev2 = rlookup loc s2
  case (ev1, ev2) of
    (Con v1, Con v2) -> deepEqV v1 v2 s1 s2
    (Abs v1, Abs v2) -> do
      cmp <- deepEqV v1 v2 s1 s2
      case cmp of
        True -> return True
        False -> do
          update loc Top
          return True
    (Top, Top) -> return True
    (Top, Abs _) -> return True
    (Abs _, Top) -> return True
    (Top, Bottom) -> return True
    (Bottom, Top) -> return True
    (Bottom, Abs _) -> return True
    (Abs _, Bottom) -> return True
    otherwise -> return False

foldlM :: Monad m => (a -> b -> m a) -> a -> [b] -> m a
foldlM f v (x:xs) = (f v x) >>= \a -> foldlM f a xs
foldlM _ v [] = return v

andM :: Monad m => Bool -> Bool -> m Bool
andM b1 b2 = do
  return (b1 && b2)

deepEqEnv :: Env -> Store -> Store -> State (Prog, Store, NameMap) Bool
deepEqEnv env s1 s2 = do
  ls <- mapM (\(_, loc) -> eqLoc loc s1 s2) env
  foldlM (\z x -> andM z x) True ls

instance Eq V where
  Prim p1 == Prim p2        = p1 == p2
  Obj c1 e1 == Obj c2 e2    = (c1==c2) && (e1==e2)
  _ == _ = False

instance Eq Primitive where
  Null == Null              = True
  StringVal s1 == StringVal s2    = s1 == s2
  IntVal i1 == IntVal i2          = i1 == i2
  BoolVal b1 == BoolVal b2        = b1 == b2
  ListVal l1 == ListVal l2        = l1 == l2
  _ == _ = False


instance Eq EV where
  Abs v1 == Abs v2 = v1 == v2
  Con v1 == Con v2 = v1 == v2
  _ == _ = False

instance Eq NameKey where
  ClassKey c1 b1 == ClassKey c2 b2            = (c1 == c2) && (b1==b2)
  MethodKey m1 n1 b1 == MethodKey m2 n2 b2    = (m1==m2) && (b1==b2) && (n1 == n2)
  _ == _                                       = False

instance Eq Modifier where
  ClassMod == ClassMod = True
  InstanceMod ==  InstanceMod = True
  _ == _ = False

compute :: Op -> Primitive -> Primitive -> Primitive
compute op v1 v2 = case op of
      Add -> (IntVal (fromInt v1 + fromInt v2))
      Sub -> (IntVal (fromInt v1 - fromInt v2))
      Mul -> (IntVal (fromInt v1 * fromInt v2))
      Lt -> (BoolVal (fromInt v1 < fromInt v2))
      Gt -> (BoolVal (fromInt v1 > fromInt v2))
      Div -> (IntVal (fromInt v1 `div` fromInt v2))
      Mod -> (IntVal (fromInt v1 `mod` fromInt v2))
      Eq -> (BoolVal (fromInt v1 == fromInt v2))
      NEq -> (BoolVal (not (fromInt v1 == fromInt v2)))

fromInt (IntVal n) = n
fromBool (BoolVal b) = b

addMethods _ [] = return ()
addMethods c (m@(MethodDef q n f b):ms) = do
    addMethod c (MethodDef q n f b)
    addMethods c ms

addMethod c method = do
    ClassDef c xs fields ctor methods <- findClass c
    replaceClass (ClassDef c xs fields ctor (method:methods))

deleteMethod c method = do
    ClassDef c xs fields ctor methods <- findClass c
    replaceClass (ClassDef c xs fields ctor (delete method methods))

replaceClass cdef@(ClassDef c _ _ _ _) = do
    deleteClass c
    addClass cdef

addClass cdef = do
    (prog, store, names) <- get
    put (cdef:prog, store, names)

deleteClass c = do
    (prog, store, names) <- get
    put (filter (\(ClassDef c' _ _ _ _)->c' /= c) prog, store, names)

bind x v = [(x, v)]

checkStore env f e2 e3 = do
      (p, m, n) <- get
      R e2'  p2 <- peval env e2
      (_, st1, _) <- get
      put (p, m, n)
      R e3' p3 <- peval env e3
      (_, st2, _) <- get
      cmp <- deepEqEnv env st1 st2
      if cmp
        then return (mAbs (f e2' e3') Bottom)
        else inconsistentChange env st1 st2

specializeAll env m argResults = do
  (_, st, _) <- get
  cmeths <- findAllMethods m (length argResults)
  when ((length cmeths) == 0) (do error ("There is no such method:"++m))
  stms <- (mapM (\(c, mdef) -> do
        (program, _, namemap) <- get
        put (program, st, namemap) 
        doCallAll (c, mdef) env argResults) cmeths)
  let m' = checkStores stms
  return m'

doCallAll (c, mdef) env argResults = do
  let (MethodDef modifier m xs body) = mdef
  (memoized, m', _) <- findMemoCall c m (projectValue argResults)
  when(not memoized) ( do
      varxs <- allocate (zip xs (projectValue argResults))
      ClassDef _ _ fs _ _ <- findClass c
      env' <- allocate (zip fs (infinite Bottom))
      varthis <- allocate (bind "this" (Abs (Obj c env')))
      R body' p <- peval (varxs++varthis++env'++env) body
      let varsxd = getRuntimeNames xs argResults
      addMethod c (MethodDef modifier m' varsxd body'))
  (_, st, _) <- get
  return (st, m')

checkStores ((st, m):sts) = m

inconsistentChange env st1 st2 =
    error ("Inconsistent change!\n" ++ show (envStore env st1) ++ "\n------\n" ++ show (envStore env st2) ++ "\n")

fieldsToEnv l = l
getStaticFields l = l

forEach :: Monad m => [a] -> (a -> m b) -> m b
forEach (x:xs) f = f x

nulls = (Null : nulls)
toCon lv = [(Con v) | v <- lv]


emptyEnv = []
list a = [a]
getNames dynEnv = [ x | (x, _) <- dynEnv]
getValues dynEnv = [ e | (_, e) <- dynEnv]
getExprs pes = [e | R e _ <- pes]

getRuntimeExprs [] = []
getRuntimeExprs ((R e (Abs _)):rs) = e:(getRuntimeExprs rs)
getRuntimeExprs ((R e Top):rs) = e:(getRuntimeExprs rs)
getRuntimeExprs ((R e Bottom):rs) = e:(getRuntimeExprs rs)
getRuntimeExprs ((R e _):rs) = getRuntimeExprs rs

getRuntimeNames [] [] = []
getRuntimeNames (x:xs) ((R e (Abs _)):rs) = x:(getRuntimeNames xs rs)
getRuntimeNames (x:xs) ((R e Top):rs) = x:(getRuntimeNames xs rs)
getRuntimeNames (x:xs) ((R e Bottom):rs) = x:(getRuntimeNames xs rs)
getRuntimeNames (x:xs) ((R e _):rs) = getRuntimeNames xs rs

makeSeq (Value _) e = e
makeSeq e (Value _) = e
makeSeq e1 e2 = Seq e1 e2


projectValue argResults = [v | (R _ v) <- argResults]

-- partition :: [X] -> [(E, EV)] -> ([(X,EV)], [(X, E)])
partition = partitionN 1
partitionN n [] [] = ([], [])
partitionN n (x:xs) ((R _ (Con v)):es) = (((x, Con v):s), d)
    where (s, d) = partitionN (n+1) xs es
partitionN n (x:xs) ((R e (Abs v)):es) = (s, ((x, e):d))
    where (s, d) = partitionN (n+1) xs es
partitionN n (x:xs) ((R e Top):es) = (s, ((x, e):d))
    where (s, d) = partitionN (n+1) xs es
partitionN n (x:xs) ((R e Bottom):es) = (s, ((x, e):d))
    where (s, d) = partitionN (n+1) xs es

getRuntimeFields fs env = filterM (isRuntime env) fs

isRuntime env f = do
  (_, st, _) <- get
  case rlookup (rlookup f env) st of
    Abs _ -> return True
    Top -> return True
    Bottom -> return True
    otherwise -> return False


toValues l = [ v | Value v <- l]

toAbs (Con x) = Abs x
toAbs v = v

mCon v = R (Value v) (Con v)
mAbs k p = R k p
envStore env st = [ (x, rlookup loc st) | (x,loc) <- env ]

trim = takeWhile (\c-> not ((c == ';') || (c == '\n')))
emptyList = []


isCompileTime (Abs _) = False
isCompileTime Top = False
isCompileTime Bottom = False  -- FIXME: make sure this is correct
isCompileTime _ = True


getValue (Con v) = v
getValue (Abs v) = v

rlookup key hash =
    case lookup key hash of
        Nothing -> error ("Undefined var/location " ++ show key ++ " in\n" ++ show hash)
        Just a -> a


classNamed c (ClassDef c' _ _ _ _) = c == c'
methodNamed m nargs (MethodDef _ m' args _) = m == m' && (nargs == (length args))

findClass c | trace ("LOOKING FOR CLASS: " ++ c) False = undefined
findClass c = do
    (prog, _, _) <- get
    case find (classNamed c) prog of
      Just cdef -> return cdef
      Nothing -> error ("missing class: " ++ c)

findMethod c m nargs | trace ("LOOKING FOR " ++ c ++ "::" ++ m ++ "::" ++ (show nargs)) False = undefined
findMethod c m nargs = do
  ClassDef _ _ _ _ methods <- findClass c
  case find (methodNamed m nargs) methods of
      Just mdef -> return mdef
      Nothing -> error ("Method not found: " ++ m)

getMethods cdef m nargs = do
  let (ClassDef c _ _ _ methods) = cdef
  case find (methodNamed m nargs) methods of
      Just mdef -> [(c, mdef)]
      Nothing -> []

findAllMethods m nargs = do
  (prog, _, _) <- get
  let lm = (aux prog [])
  if (length lm) > 0 then do
    return lm
   else error ("Method not found: " ++ m)
    where
      aux [] l = l
      aux (c:cs) l = aux cs (l++(getMethods c m nargs))

freevars :: C -> [M] -> E -> State (Prog, Store, NameMap) [X]
freevars c lm e = do
  fvaux e [] lm
  where
    -- fvaux search for free variables in e and it
    -- carries the bound variables in the second arg
    -- the third argument is a list of method visited
    fvaux :: E -> [X] -> [M] -> State (Prog, Store, NameMap) [X]
    fvaux (Value _) l _ = return []
    fvaux (Var x) l _ = return ( if elem x l then [] else [x])
    fvaux (Bop _ e1 e2) l lm = do
      f1 <- (fvaux e1 l lm)
      f2 <- (fvaux e2 l lm)
      return (f1 ++ f2)
    fvaux (Assign x e) l lm = do
      let l1 = if elem x l then [] else [x]
      l2 <- (fvaux e l lm)
      return (l1++l2)
    fvaux (If e1 e2 e3) l lm = do
      l1 <- (fvaux e1 l lm)
      l2 <- (fvaux e2 l lm)
      l3 <- (fvaux e3 l lm)
      return (l1 ++ l2 ++ l3)
    fvaux (Seq e1 e2) l lm = do
      l1 <- (fvaux e1 l lm)
      l2 <- (fvaux e2 l lm)
      return (l1 ++ l2)
    fvaux (Let x e1 e2) l lm = do
      l1 <- (fvaux e1 (x:l) lm)
      l2 <- (fvaux e2 (x:l) lm)
      return (l1 ++ l2)
    fvaux (While e1 e2) l lm = do
      l1 <- (fvaux e1 l lm)
      l2 <- (fvaux e2 l lm)
      return (l1 ++ l2)
    fvaux (ClassVal _) l _ = return []
    fvaux (Call e m a) l lm =
      case e of
        Var "this" -> do
          MethodDef _ mn _ b <- findMethod c m (length a)
          if elem mn lm then do
             fvaux e l lm
           else do
             l1 <- (freevars c lm b)
             l2 <- (fvaux e l lm)
             return (l1 ++ l2)
          --return []
        otherwise -> do
             l1 <- fvaux e l lm
             l2 <- fvauxList a l lm
             return (l1 ++ l2)
    fvaux (New c args) l lm = fvauxList args l lm

    fvauxList [] l _ = return []
    fvauxList (y:ys) l lm = do
      l1 <- fvaux y l lm
      l2 <- fvauxList ys l lm
      return (l1++l2)

-- -- parser -- --

ident = do
    x <- sat isAlpha
    xs <- many alphanum
    many (sat isSpace)
    return (x:xs)

pStatList = do
    stats <- many pStat
    if stats == [] then
      return (Value (Prim Null))
     else
      return (foldl1 Seq stats)
pStat = pLet +++ pIf +++ pWhile +++ pGroup +++ pExpStat

pExpStat = do
    e <- pExp
    symbol ";"
    return e

pGroup = do
    symbol "{"
    e <- pStatList
    symbol "}"
    return e

pExp = pNew +++ (pExp2 `chainl1` pRelOp)
pExp2 = pExp3 `chainl1` pAddOp
pExp3 = pCall `chainl1` pMulOp

pMulOp =  ops [ (symbol "%", Bop Mod),  (symbol "*", Bop Mul), (symbol "/", Bop Div)]
pAddOp =  ops [ (symbol "+", Bop Add), (symbol "-", Bop Sub)]
pRelOp = ops [ (symbol "=", Bop Eq), (symbol "!=", Bop NEq), (symbol "<", Bop Lt), (symbol ">", Bop Gt)]

cont :: a -> (a -> Parser a) -> Parser a
cont x f = (do { z <- f x;  cont z f }) +++ (return x)

pCall = do
    e <- pInvoke
    cont e (\e -> do
     symbol "."
     m <- ident
     a <- args
     return $ Call e m a)

pInvoke = do
    t <- pBase
    cont t (\t-> do
     symbol "@"
     m <- pExp
     a <- args
     return $ Invoke t m a)

pLet = do
    (x, e) <- pVarDef
    b <- pStatList
    return $ Let x e b

pIf = do
    symbol "if"
    a <- pParen
    b <- pStat
    c <- pElse +++ (return (Value (Prim Null)))
    return $ If a b c

pElse = do
  symbol "else"
  pStat

pWhile = do
    symbol "while"
    a <- pParen
    b <- pStat
    return $ While a b

-- pFor = do
--     symbol "for"
--     x <- ident
--     symbol ":"
--     a <- pExp
--     symbol "do"
--     b <- pStat
--     return $ For x a b

--- pInstanceOf a c = a symbol " instanceof" c

--- pClassVal c = c
--- p(ClassVal c) = c

-- pCast c a = do
--     symbol "("
--     c <- ident
--     symbol ")"
--     a <- args

pNew = do
    symbol "new"
    c <- ident
    a <- args
    return $ New c a


args = do
    symbol "("
    a <- pExp `sepby` (symbol ",")
    symbol ")"
    return $ a

pFormals = do
    symbol "("
    a <- ident `sepby` (symbol ",")
    symbol ")"
    return $ a

pBase = pCT +++ pRT +++ pValue  +++ pParen +++ pVar

pParen = do
    symbol "("
    e <- pExp
    symbol ")"
    return e

pCT = do
    symbol "CT"
    symbol "("
    e <- pExp
    c <- optCond
    symbol ")"
    return (CT e c)

optCond = do { symbol ","; pExp +++ pISCT} +++ (return (Value (Prim (BoolVal True))))

pRT = do
    symbol "RT"
    symbol "("
    e <- pExp
    symbol ")"
    return (RT e)

{-
pISKNOWN = do
    symbol "ISKNOWN"
    symbol "("
    e <- pExp
    symbol ")"
    return (IsKnown e)
-}

pISCT = do
    symbol "ISCT"
    symbol "("
    e <- pExp
    symbol ")"
    return (IsCT e)

pVar = do
    x <- ident
    cont (Var x) pVar2
pVar2 (Var x) = do
     symbol ":="
     a <- pExp
     return $ Assign x a
pVar2 e = mzero

pValue = do
    pInt +++ pString +++ pConst

pConst =
    ops [(symbol "true", Value (Prim (BoolVal True))),
            (symbol "false", Value (Prim (BoolVal False))),
        (symbol "null", Value (Prim Null)),
        (symbol "skip", Value (Prim Null))]

pString = do
    s <- bracket (char '"') (many item) (char '"')
    return $ Value (Prim (StringVal s))

pInt = do
    n <- integer
    return $ Value (Prim (IntVal n))

pClass = do
    symbol "class"
    c <- ident
    args <- pFormals
    symbol "{"
    fields <- many pField
    ctor <- pInit +++ (return (Value (Prim Null)))
    methods <- many pMethod
    symbol "}"
    return $ ClassDef c args fields ctor methods

pInit = do
    symbol "init"
    body <- pGroup
    return body

pField = do
    symbol "var"
    x <- ident
    symbol ";"
    return x

pVarDef = do
    symbol "var"
    x <- ident
    symbol "="
    e <- pExp
    symbol ";"
    return (x, e)

pModifier =
    ops [(symbol "static", ClassMod), (symbol "method", InstanceMod)]

pMethod = do
    mod <- pModifier
    m <- ident
    args <- pFormals
    body <- pGroup
    return $ MethodDef mod m args body

pMool = many pClass

parse p s =
    case papply p s of
        [] -> error ("No parse of " ++ s)
        [(e,"")] -> e
        ((e,s):_) -> error ("parse error at " ++ s)

test prog cmd =
  let (v,(p,m,n)) = Control.Monad.State.runState (peval [] cmd) (prog, [], []) in do
    print cmd
    putStr "===result===\n"
    print v
    putStr "Store: "
    print m
    --putStr "\nEnv:"
    --print p
    --putStr "\n"
--    print ss
--    putStr "\n"

pevalAllMethods cdef@(ClassDef _ _ _ _ methods) =
    mapM (pevalTopMethod cdef) methods

pevalTopMethod (ClassDef c xs fields _ _) meth@(MethodDef modifier m args b)  = do
    env <- allocate ([("this",  Abs (Obj c []))] ++ [ (x, Top) | x <- fields ])
    meth' <- pevalm env meth
    deleteMethod c meth
    addMethods c [meth']

runFile f = do
    putStr ("\n**** READING " ++ f ++ "\n");
    src <- readFile f
    let prog = parse pMool src
    -- putStr (showList "\n\n" prog) -- print the file
    let (_,(p,_,_)) = Control.Monad.State.runState (mapM pevalAllMethods prog) (prog,[],[])
    writeFile (f++".out") ((showlist "\n\n" p)++"\n")

testFile f =
  catch (runFile f)
         (\e -> do let err = show (e :: SomeException)
                   putStr ("Error: " ++ err)
                   writeFile (f++".out") ("Error: " ++ err)
                   return ())

fevalMain c = do
  MethodDef _ _ xs body <- findMethod c "main" 0
  feval [] body

pevalMain c = do
  meth<- findMethod c "main" 0
  meth' <- pevalm [] meth
  deleteMethod c meth
  addMethods c [meth']

runFileMain f c = do
    putStr ("\n**** READING " ++ f ++ "\n");
    src <- readFile f
    let prog = parse pMool src
    let (_,(p,_,_)) = Control.Monad.State.runState (pevalMain c) (prog,[],[])
    writeFile (f++".out") ((showlist "\n\n" p)++"\n")

testFileMain f c =
  catch (runFileMain f c)
         (\e -> do let err = show (e :: SomeException)
                   putStr ("Error: " ++ err)
                   writeFile (f++".out") ("Error: " ++ err)
                   return ())


runFevalFile f c = do
    putStr ("\n**** READING " ++ f ++ "\n");
    src <- readFile f
    let prog = parse pMool src
    let (v,(_,_,_)) = Control.Monad.State.runState (fevalMain c) (prog,[],[])
    putStr ("Output: " ++ (show v) ++ "\n")

fevalFile f c = 
  catch (runFevalFile f c)
         (\e -> do let err = show (e :: SomeException)
                   putStr ("Error: " ++ err)
                   writeFile (f++".out") ("Error: " ++ err)
                   return ())


main = do
  testFile "tests/t1.mool"
  testFile "tests/t2.mool"
  testFile "tests/t3.mool"
  testFile "tests/t4.mool"
  testFile "tests/t5.mool"
  testFile "tests/t6.mool"
  testFile "tests/t7.mool"
  testFile "tests/t8.mool"
--  fevalFile "tests/t9.mool" "Main"
--  testFileMain "tests/t9.mool" "Main"


